import os.path
from typing import Iterable, List, Sequence, Tuple

from aiogram import Router
from aiogram.fsm.state import State
from diagrams import Cluster, Diagram, Edge
from diagrams.custom import Custom

from aiogram_dialog import Dialog
from aiogram_dialog.api.internal import WindowProtocol
from aiogram_dialog.manager.setup import collect_dialogs
from aiogram_dialog.widgets.kbd import (
    Back,
    Cancel,
    Group,
    Next,
    Start,
    SwitchTo,
)

ICON_PATH = os.path.join(os.path.dirname(__file__), "calculator.png")


def widget_edges(nodes, dialog, starts, current_state, kbd):
    states = dialog.states()
    if isinstance(kbd, Start):
        nodes[current_state] >> Edge(color="#338a3e") >> nodes[kbd.state]
    elif isinstance(kbd, SwitchTo):
        nodes[current_state] >> Edge(color="#0086c3") >> nodes[kbd.state]
    elif isinstance(kbd, Next):
        new_state = states[states.index(current_state) + 1]
        nodes[current_state] >> Edge(color="#0086c3") >> nodes[new_state]
    elif isinstance(kbd, Back):
        new_state = states[states.index(current_state) - 1]
        nodes[current_state] >> Edge(color="grey") >> nodes[new_state]
    elif isinstance(kbd, Cancel):
        for from_, to_ in starts:
            if to_.group == current_state.group:
                (
                    nodes[current_state] >>
                    Edge(color="grey", style="dashed") >>
                    nodes[from_]
                )


def walk_keyboard(
        nodes,
        dialog,
        starts: List[Tuple[State, State]],
        current_state: State,
        keyboards: Sequence,
):
    for kbd in keyboards:
        if isinstance(kbd, Group):
            walk_keyboard(nodes, dialog, starts, current_state, kbd.buttons)
        else:
            widget_edges(nodes, dialog, starts, current_state, kbd)


def find_starts(
        current_state, keyboards: Sequence,
) -> Iterable[Tuple[State, State]]:
    for kbd in keyboards:
        if isinstance(kbd, Group):
            yield from find_starts(current_state, kbd.buttons)
        elif isinstance(kbd, Start):
            yield current_state, kbd.state


def render_window(
        nodes: dict, dialog: Dialog, starts: List[Tuple[State, State]],
        window: WindowProtocol,
):
    walk_keyboard(
        nodes,
        dialog,
        starts,
        window.get_state(),
        [window.keyboard],
    )
    preview_add_transitions = getattr(
        window, "preview_add_transitions", None,
    )
    if preview_add_transitions:
        walk_keyboard(
            nodes,
            dialog,
            starts,
            window.get_state(),
            preview_add_transitions,
        )


def render_transitions(
        router: Router,
        title: str = "Aiogram Dialog",
        filename: str = "aiogram_dialog",
        format: str = "png",
):
    dialogs = list(collect_dialogs(router))
    with Diagram(title, filename=filename, outformat=format, show=False):
        nodes = {}
        for dialog in dialogs:
            with Cluster(dialog.states_group_name()):
                for window in dialog.windows.values():
                    nodes[window.get_state()] = Custom(
                        icon_path=ICON_PATH, label=window.get_state().state,
                    )

        starts = []
        for dialog in dialogs:
            for window in dialog.windows.values():
                starts.extend(
                    find_starts(window.get_state(), [window.keyboard]),
                )

        for dialog in dialogs:
            for window in dialog.windows.values():
                render_window(
                    nodes=nodes, dialog=dialog, window=window, starts=starts,
                )



import re
from typing import Any, Optional

from aiogram_dialog.api.exceptions import InvalidWidgetIdError
from aiogram_dialog.api.protocols import DialogManager
from .base import BaseWidget

ID_PATTERN = re.compile("^[a-zA-Z0-9_.]+$")


class Actionable(BaseWidget):
    def __init__(self, id: Optional[str] = None):
        if id and not ID_PATTERN.match(id):
            raise InvalidWidgetIdError(f"Invalid widget id: {id}")
        self.widget_id = id

    def find(self, widget_id: str):
        """Find nested widget or current one by id."""
        if self.widget_id is not None and self.widget_id == widget_id:
            return self
        return None

    def get_widget_data(
            self,
            manager: DialogManager,
            default: Any,
    ) -> Any:
        """Get data for current widget id, setting default if needed."""
        return manager.current_context().widget_data.setdefault(
            self.widget_id,
            default,
        )

    def set_widget_data(
            self,
            manager: DialogManager,
            value: Any,
    ) -> Any:
        """Set data for current widget id."""
        manager.current_context().widget_data[self.widget_id] = value



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from aiogram import Bot
from jinja2 import BaseLoader, Environment

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from .base import Text

BOT_ENV_FIELD = "DialogsJinjaEnvironment"

Filter = Callable[..., str]
Filters = Union[Iterable[Tuple[str, Filter]], Mapping[str, Filter]]


class Jinja(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        super().__init__(when=when)
        self.template_text = text

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        bot: Bot = manager.middleware_data["bot"]
        env: Environment = getattr(bot, BOT_ENV_FIELD, default_env)
        template = env.get_template(self.template_text)

        if env.is_async:
            return await template.render_async(data)
        else:
            return template.render(data)


class StubLoader(BaseLoader):
    def get_source(self, environment, template):
        del environment  # unused
        return template, template, lambda: True


def _create_env(
        *args: Any, filters: Optional[Filters] = None, **kwargs: Any,
) -> Environment:
    kwargs.setdefault("autoescape", True)
    kwargs.setdefault("lstrip_blocks", True)
    kwargs.setdefault("trim_blocks", True)
    if "loader" not in kwargs:
        kwargs["loader"] = StubLoader()
    env = Environment(*args, **kwargs)
    if filters is not None:
        env.filters.update(filters)
    return env


def setup_jinja(
        bot: Bot,
        *args: Any,
        filters: Optional[Filters] = None,
        **kwargs: Any,
) -> Environment:
    env = _create_env(*args, filters=filters, **kwargs)
    setattr(bot, BOT_ENV_FIELD, env)
    return env


default_env = _create_env()



from typing import Any, Callable, Dict, List, Optional, Union

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.internal import Widget
from aiogram_dialog.api.protocols import (
    DialogManager, DialogProtocol,
)
from aiogram_dialog.manager.sub_manager import SubManager
from aiogram_dialog.widgets.common import ManagedWidget, WhenCondition
from .base import Keyboard
from ..common.items import get_items_getter, ItemsGetterVariant

ItemIdGetter = Callable[[Any], Union[str, int]]


class ListGroup(Keyboard):
    def __init__(
            self,
            *buttons: Keyboard,
            id: Optional[str] = None,
            item_id_getter: ItemIdGetter,
            items: ItemsGetterVariant,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        self.buttons = buttons
        self.item_id_getter = item_id_getter
        self.items_getter = get_items_getter(items)

    async def _render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        kbd: List[List[InlineKeyboardButton]] = []
        for pos, item in enumerate(self.items_getter(data)):
            kbd.extend(await self._render_item(pos, item, data, manager))
        return kbd

    async def _render_item(
            self,
            pos: int,
            item: Any,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        kbd: List[List[InlineKeyboardButton]] = []
        data = {"data": data, "item": item, "pos": pos + 1, "pos0": pos}
        item_id = str(self.item_id_getter(item))
        sub_manager = SubManager(
            widget=self,
            manager=manager,
            widget_id=self.widget_id,
            item_id=item_id,
        )
        for b in self.buttons:
            b_kbd = await b.render_keyboard(data, sub_manager)
            for row in b_kbd:
                for btn in row:
                    if btn.callback_data:
                        btn.callback_data = self._item_callback_data(
                            f"{item_id}:{btn.callback_data}",
                        )
            kbd.extend(b_kbd)
        return kbd

    def find(self, widget_id: str) -> Optional[Widget]:
        if widget_id == self.widget_id:
            return self
        for btn in self.buttons:
            widget = btn.find(widget_id)
            if widget:
                return widget
        return None

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        item_id, callback_data = data.split(":", maxsplit=1)
        callback = callback.model_copy(update={
            "data": callback_data,
        })
        sub_manager = SubManager(
            widget=self,
            manager=manager,
            widget_id=self.widget_id,
            item_id=item_id,
        )
        for b in self.buttons:
            if await b.process_callback(callback, dialog, sub_manager):
                return True

    def managed(self, manager: DialogManager) -> "ManagedListGroup":
        return ManagedListGroup(self, manager)


class ManagedListGroup(ManagedWidget[ListGroup]):
    def find_for_item(self, widget_id: str, item_id: str) -> Optional[Any]:
        widget = self.widget.find(widget_id)
        if widget:
            return widget.managed(
                SubManager(
                    widget=self.widget,
                    manager=self.manager,
                    widget_id=self.widget.widget_id,
                    item_id=item_id,
                ),
            )
        return None



__all__ = [
    "FakeChat", "FakeUser",
    "DialogManagerFactory",
    "CALLBACK_DATA_KEY", "CONTEXT_KEY", "STACK_KEY", "STORAGE_KEY",
    "DataGetter", "InputWidget", "KeyboardWidget",
    "MediaWidget", "TextWidget", "Widget",
    "WindowProtocol",
]

from .fake_data import FakeChat, FakeUser
from .manager import (
    DialogManagerFactory,
)
from .middleware import (
    CALLBACK_DATA_KEY, CONTEXT_KEY, STACK_KEY, STORAGE_KEY,
)
from .widgets import (
    DataGetter, InputWidget, KeyboardWidget,
    MediaWidget, TextWidget, Widget,
)
from .window import WindowProtocol



from typing import Dict, List, Optional

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import (
    BaseScroll, OnPageChangedVariants, WhenCondition,
)
from .base import Keyboard
from .group import Group


class ScrollingGroup(Group, BaseScroll):
    def __init__(
            self,
            *buttons: Keyboard,
            id: str,
            width: Optional[int] = None,
            height: int = 0,
            when: WhenCondition = None,
            on_page_changed: OnPageChangedVariants = None,
            hide_on_single_page: bool = False,
            hide_pager: bool = False,
    ):
        Group.__init__(self, *buttons, id=id, width=width, when=when)
        BaseScroll.__init__(self, id=id, on_page_changed=on_page_changed)
        self.height = height
        self.hide_on_single_page = hide_on_single_page
        self.hide_pager = hide_pager

    def _get_page_count(
            self,
            keyboard: List[List[InlineKeyboardButton]],
    ) -> int:
        return len(keyboard) // self.height + bool(len(keyboard) % self.height)

    async def _render_contents(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return await super()._render_keyboard(data, manager)

    async def _render_pager(
            self,
            pages: int,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        if self.hide_pager:
            return []
        if pages == 0 or (pages == 1 and self.hide_on_single_page):
            return []

        last_page = pages - 1
        current_page = min(last_page, await self.get_page(manager))
        next_page = min(last_page, current_page + 1)
        prev_page = max(0, current_page - 1)

        return [
            [
                InlineKeyboardButton(
                    text="1", callback_data=self._item_callback_data("0"),
                ),
                InlineKeyboardButton(
                    text="<",
                    callback_data=self._item_callback_data(prev_page),
                ),
                InlineKeyboardButton(
                    text=str(current_page + 1),
                    callback_data=self._item_callback_data(current_page),
                ),
                InlineKeyboardButton(
                    text=">",
                    callback_data=self._item_callback_data(next_page),
                ),
                InlineKeyboardButton(
                    text=str(last_page + 1),
                    callback_data=self._item_callback_data(last_page),
                ),
            ],
        ]

    async def _render_page(
            self,
            page: int,
            keyboard: List[List[InlineKeyboardButton]],
    ) -> List[List[InlineKeyboardButton]]:
        pages = self._get_page_count(keyboard)
        last_page = pages - 1
        current_page = min(last_page, page)
        page_offset = current_page * self.height

        return keyboard[page_offset: page_offset + self.height]

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        keyboard = await self._render_contents(data, manager)
        pages = self._get_page_count(keyboard)

        pager = await self._render_pager(pages, manager)
        page_keyboard = await self._render_page(
            page=await self.get_page(manager),
            keyboard=keyboard,
        )

        return page_keyboard + pager

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        await self.set_page(callback, int(data), manager)
        return True

    async def get_page_count(self, data: Dict, manager: DialogManager) -> int:
        keyboard = await self._render_contents(data, manager)
        return self._get_page_count(keyboard=keyboard)



from enum import Enum


class ShowMode(Enum):
    AUTO = "auto"
    EDIT = "edit"
    SEND = "send"


class StartMode(Enum):
    NORMAL = "NORMAL"
    RESET_STACK = "RESET_STACK"
    NEW_STACK = "NEW_STACK"



from typing import Dict

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from .base import Text
from ..common.items import get_items_getter, ItemsGetterVariant


class List(Text):
    def __init__(
            self,
            field: Text,
            items: ItemsGetterVariant,
            sep: str = "\n",
            when: WhenCondition = None,
    ):
        super().__init__(when=when)
        self.field = field
        self.sep = sep
        self.items_getter = get_items_getter(items)

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        texts = [
            await self.field.render_text(
                {"data": data, "item": item, "pos": pos + 1, "pos0": pos},
                manager,
            )
            for pos, item in enumerate(self.items_getter(data))
        ]
        return self.sep.join(filter(None, texts))



from logging import getLogger
from typing import Dict, List, Optional

from aiogram.fsm.state import State
from aiogram.types import (
    CallbackQuery,
    InlineKeyboardMarkup,
    Message,
    UNSET_PARSE_MODE,
)

from aiogram_dialog.api.entities import MediaAttachment, NewMessage
from aiogram_dialog.api.internal import Widget, WindowProtocol
from .api.protocols import DialogManager, DialogProtocol
from .widgets.data import PreviewAwareGetter
from .widgets.kbd import Keyboard
from .widgets.utils import (
    ensure_data_getter,
    ensure_widgets,
    GetterVariant,
    WidgetSrc,
)

logger = getLogger(__name__)


class Window(WindowProtocol):
    def __init__(
            self,
            *widgets: WidgetSrc,
            state: State,
            getter: GetterVariant = None,
            parse_mode: Optional[str] = UNSET_PARSE_MODE,
            disable_web_page_preview: Optional[bool] = None,
            preview_add_transitions: Optional[List[Keyboard]] = None,
            preview_data: GetterVariant = None,
    ):
        (
            self.text,
            self.keyboard,
            self.on_message,
            self.media,
        ) = ensure_widgets(widgets)
        self.getter = PreviewAwareGetter(
            ensure_data_getter(getter),
            ensure_data_getter(preview_data),
        )
        self.state = state
        self.parse_mode = parse_mode
        self.disable_web_page_preview = disable_web_page_preview
        self.preview_add_transitions = preview_add_transitions

    async def render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        return await self.text.render_text(data, manager)

    async def render_media(
            self, data: Dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        if self.media:
            return await self.media.render_media(data, manager)

    async def render_kbd(
            self, data: Dict, manager: DialogManager,
    ) -> InlineKeyboardMarkup:
        keyboard = await self.keyboard.render_keyboard(data, manager)
        return InlineKeyboardMarkup(inline_keyboard=keyboard)

    async def load_data(
            self, dialog: "DialogProtocol",
            manager: DialogManager,
    ) -> Dict:
        data = await dialog.load_data(manager)
        data.update(await self.getter(**manager.middleware_data))
        return data

    async def process_message(
            self, message: Message, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> None:
        if self.on_message:
            await self.on_message.process_message(message, dialog, manager)

    async def process_callback(
            self, callback: CallbackQuery, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> None:
        if self.keyboard:
            await self.keyboard.process_callback(callback, dialog, manager)

    async def render(
            self, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> NewMessage:
        logger.debug("Show window: %s", self)
        chat = manager.middleware_data["event_chat"]
        current_data = await self.load_data(dialog, manager)
        return NewMessage(
            chat=chat,
            text=await self.render_text(current_data, manager),
            reply_markup=await self.render_kbd(current_data, manager),
            parse_mode=self.parse_mode,
            disable_web_page_preview=self.disable_web_page_preview,
            media=await self.render_media(current_data, manager),
        )

    def get_state(self) -> State:
        return self.state

    def find(self, widget_id) -> Optional[Widget]:
        for root in (self.text, self.keyboard, self.on_message):
            if root:
                if found := root.find(widget_id):
                    return found
        return None

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}({self.state})>"



from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Dict, List, Optional, Union

from aiogram.types import CallbackQuery, InlineKeyboardButton

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import ManagedWidget, WhenCondition
from aiogram_dialog.widgets.text import Case, Text
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)
from .base import Keyboard

OnStateChanged = Callable[
    [ChatEvent, "ManagedCheckbox", DialogManager], Awaitable,
]
OnStateChangedVariant = Union[
    OnStateChanged, WidgetEventProcessor, None,
]


class BaseCheckbox(Keyboard, ABC):
    def __init__(
            self,
            checked_text: Text,
            unchecked_text: Text,
            id: str,
            on_click: OnStateChangedVariant = None,
            on_state_changed: OnStateChangedVariant = None,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        self.text = Case(
            {True: checked_text, False: unchecked_text},
            selector=self._is_text_checked,
        )
        self.on_click = ensure_event_processor(on_click)
        self.on_state_changed = ensure_event_processor(on_state_changed)

    async def _render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        checked = int(self.is_checked(manager))
        # store current checked status in callback data
        return [
            [
                InlineKeyboardButton(
                    text=await self.text.render_text(data, manager),
                    callback_data=self._item_callback_data(checked),
                ),
            ],
        ]

    async def _process_item_callback(
            self,
            callback: CallbackQuery,
            data: str,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        # remove prefix and cast "0" as False, "1" as True
        checked = data != "0"
        await self.on_click.process_event(
            callback, self.managed(manager), manager,
        )
        await self.set_checked(callback, not checked, manager)
        return True

    def _is_text_checked(
            self, data: Dict, case: Case, manager: DialogManager,
    ) -> bool:
        del data  # unused
        del case  # unused
        return self.is_checked(manager)

    @abstractmethod
    def is_checked(self, manager: DialogManager) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def set_checked(
            self, event: ChatEvent, checked: bool,
            manager: DialogManager,
    ):
        raise NotImplementedError


class Checkbox(BaseCheckbox):
    def __init__(
            self,
            checked_text: Text,
            unchecked_text: Text,
            id: str,
            on_click: Union[OnStateChanged, WidgetEventProcessor, None] = None,
            on_state_changed: Optional[OnStateChanged] = None,
            default: bool = False,
            when: WhenCondition = None,
    ):
        super().__init__(
            checked_text=checked_text, unchecked_text=unchecked_text,
            on_click=on_click, on_state_changed=on_state_changed,
            id=id, when=when,
        )
        self.default = default

    def is_checked(self, manager: DialogManager) -> bool:
        return self.get_widget_data(manager, self.default)

    async def set_checked(
            self, event: ChatEvent, checked: bool,
            manager: DialogManager,
    ) -> None:
        self.set_widget_data(manager, checked)
        await self.on_state_changed.process_event(
            event, self.managed(manager), manager,
        )

    def managed(self, manager: DialogManager) -> "ManagedCheckbox":
        return ManagedCheckbox(self, manager)


class ManagedCheckbox(ManagedWidget[Checkbox]):
    def is_checked(self) -> bool:
        return self.widget.is_checked(self.manager)

    async def set_checked(self, checked: bool) -> None:
        return await self.widget.set_checked(
            self.manager.event, checked, self.manager,
        )



from abc import abstractmethod
from typing import Any, Awaitable, Callable, Optional, Sequence, Union

from aiogram import F
from aiogram.dispatcher.event.handler import FilterObject
from aiogram.types import ContentType, Message

from aiogram_dialog.api.internal import InputWidget
from aiogram_dialog.api.protocols import (
    DialogManager, DialogProtocol,
)
from aiogram_dialog.widgets.common import Actionable
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)

MessageHandlerFunc = Callable[
    [Message, "MessageInput", DialogManager],
    Awaitable,
]


class BaseInput(Actionable, InputWidget):
    @abstractmethod
    async def process_message(
            self, message: Message, dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        raise NotImplementedError


class MessageInput(BaseInput):
    def __init__(
            self,
            func: Union[MessageHandlerFunc, WidgetEventProcessor, None],
            content_types: Union[Sequence[str], str] = ContentType.ANY,
            filter: Optional[Callable[..., Any]] = None,
    ):
        super().__init__()
        self.func = ensure_event_processor(func)

        filters = []
        if isinstance(content_types, str):
            if content_types != ContentType.ANY:
                filters.append(FilterObject(F.content_type == content_types))
        else:
            if ContentType.ANY not in content_types:
                filters.append(FilterObject(F.content_type.in_(content_types)))
        if filter is not None:
            filters.append(FilterObject(filter))
        self.filters = filters

    async def process_message(
            self,
            message: Message,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        for handler_filter in self.filters:
            if not await handler_filter.call(
                    manager.event, **manager.middleware_data,
            ):
                return False
        await self.func.process_event(message, self, manager)
        return True



from typing import Any, Optional

from aiogram_dialog import DialogManager
from aiogram_dialog.api.internal import Widget


class BaseWidget(Widget):
    def managed(self, manager: DialogManager) -> Any:
        return self

    def find(self, widget_id: str) -> Optional["Widget"]:
        return None



from dataclasses import dataclass, field
from typing import Dict, List, Union

from aiogram.fsm.state import State

Data = Union[Dict, List, int, str, float, None]
DataDict = Dict[str, Data]


@dataclass(unsafe_hash=True)
class Context:
    _intent_id: str = field(compare=True)
    _stack_id: str = field(compare=True)
    state: State = field(compare=False)
    start_data: Data = field(compare=False)
    dialog_data: DataDict = field(compare=False, default_factory=dict)
    widget_data: DataDict = field(compare=False, default_factory=dict)

    @property
    def id(self) -> str:
        return self._intent_id

    @property
    def stack_id(self) -> str:
        return self._stack_id



from typing import Any, Callable, Optional

from aiogram.dispatcher.event.handler import FilterObject
from aiogram.types import Message

from aiogram_dialog.api.protocols import (
    DialogManager, DialogProtocol,
)
from .base import BaseInput


class CombinedInput(BaseInput):
    def __init__(
            self,
            *inputs: BaseInput,
            filter: Optional[Callable[..., Any]] = None,
    ):
        super().__init__()
        self.inputs = inputs
        self.filters = []
        if filter is not None:
            self.filters.append(FilterObject(filter))

    async def process_message(
            self,
            message: Message,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        for handler_filter in self.filters:
            if not await handler_filter.call(
                    manager.event, **manager.middleware_data,
            ):
                return False
        for input_widget in self.inputs:
            if await input_widget.process_message(message, dialog, manager):
                return True
        return False



from typing import Awaitable, Callable, Dict, List, Optional, Union

from aiogram.types import CallbackQuery, InlineKeyboardButton, WebAppInfo

from aiogram_dialog.api.protocols import DialogManager, DialogProtocol
from aiogram_dialog.widgets.common import WhenCondition
from aiogram_dialog.widgets.text import Text
from aiogram_dialog.widgets.widget_event import (
    ensure_event_processor,
    WidgetEventProcessor,
)
from .base import Keyboard

OnClick = Callable[[CallbackQuery, "Button", DialogManager], Awaitable]


class Button(Keyboard):
    def __init__(
            self,
            text: Text,
            id: str,
            on_click: Union[OnClick, WidgetEventProcessor, None] = None,
            when: WhenCondition = None,
    ):
        super().__init__(id=id, when=when)
        self.text = text
        self.on_click = ensure_event_processor(on_click)

    async def _process_own_callback(
            self,
            callback: CallbackQuery,
            dialog: DialogProtocol,
            manager: DialogManager,
    ) -> bool:
        await self.on_click.process_event(callback, self, manager)
        return True

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            [
                InlineKeyboardButton(
                    text=await self.text.render_text(data, manager),
                    callback_data=self._own_callback_data(),
                ),
            ],
        ]


class Url(Keyboard):
    def __init__(
            self,
            text: Text,
            url: Text,
            id: Optional[str] = None,
            when: Union[str, Callable, None] = None,
    ):
        super().__init__(id=id, when=when)
        self.text = text
        self.url = url

    async def _render_keyboard(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            [
                InlineKeyboardButton(
                    text=await self.text.render_text(data, manager),
                    url=await self.url.render_text(data, manager),
                ),
            ],
        ]


class WebApp(Url):
    async def _render_keyboard(
            self, data: Dict, manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        text = await self.text.render_text(data, manager)

        web_app_url = await self.url.render_text(data, manager)
        web_app_info = WebAppInfo(url=web_app_url)

        return [[InlineKeyboardButton(text=text, web_app=web_app_info)]]


class SwitchInlineQuery(Keyboard):
    def __init__(
        self,
        text: Text,
        switch_inline_query: Text,
        id: Optional[str] = None,
        when: Union[str, Callable, None] = None,
    ):
        super().__init__(id=id, when=when)
        self.text = text
        self.switch_inline = switch_inline_query

    async def _render_keyboard(
        self,
        data: Dict,
        manager: DialogManager,
    ) -> List[List[InlineKeyboardButton]]:
        return [
            [
                InlineKeyboardButton(
                    text=await self.text.render_text(data, manager),
                    switch_inline_query=await self.switch_inline.render_text(
                        data, manager,
                    ),
                ),
            ],
        ]



from typing import Any, Dict
from unittest.mock import Mock

import pytest
from aiogram import Dispatcher
from aiogram.filters import CommandStart
from aiogram.fsm.state import State, StatesGroup
from aiogram.fsm.storage.memory import MemoryStorage
from aiogram.types import Message

from aiogram_dialog import (
    Dialog, DialogManager, setup_dialogs, StartMode, Window,
)
from aiogram_dialog.test_tools import BotClient, MockMessageManager
from aiogram_dialog.test_tools.keyboard import InlineButtonTextLocator
from aiogram_dialog.widgets.kbd import Button
from aiogram_dialog.widgets.text import Const, Format


class MainSG(StatesGroup):
    start = State()
    next = State()


async def on_click(event, button, manager: DialogManager) -> None:
    manager.middleware_data["usecase"]()
    await manager.next()


async def on_finish(event, button, manager: DialogManager) -> None:
    await manager.done()


async def second_getter(user_getter, **kwargs) -> Dict[str, Any]:
    return {
        "user": user_getter(),
    }


dialog = Dialog(
    Window(
        Format("stub"),
        Button(Const("Button"), id="hello", on_click=on_click),
        state=MainSG.start,
    ),
    Window(
        Format("Next {user}"),
        Button(Const("Finish"), id="hello", on_click=on_finish),
        state=MainSG.next,
        getter=second_getter,
    ),
)


async def start(message: Message, dialog_manager: DialogManager):
    await dialog_manager.start(MainSG.start, mode=StartMode.RESET_STACK)


@pytest.mark.asyncio
async def test_click():
    usecase = Mock()
    user_getter = Mock(side_effect=["Username"])
    dp = Dispatcher(
        usecase=usecase, user_getter=user_getter,
        storage=MemoryStorage(),
    )
    dp.include_router(dialog)
    dp.message.register(start, CommandStart())

    client = BotClient(dp)
    message_manager = MockMessageManager()
    setup_dialogs(dp, message_manager=message_manager)

    # start
    await client.send("/start")
    first_message = message_manager.one_message()
    assert first_message.text == "stub"
    assert first_message.reply_markup
    user_getter.assert_not_called()

    # redraw
    message_manager.reset_history()
    await client.send("whatever")

    first_message = message_manager.one_message()
    assert first_message.text == "stub"

    # click next
    message_manager.reset_history()
    callback_id = await client.click(
        first_message, InlineButtonTextLocator("Button"),
    )

    message_manager.assert_answered(callback_id)
    usecase.assert_called()
    second_message = message_manager.one_message()
    assert second_message.text == "Next Username"
    assert second_message.reply_markup.inline_keyboard
    user_getter.assert_called_once()



from .api.entities import (
    ChatEvent, Data, DEFAULT_STACK_ID, LaunchMode, ShowMode, StartMode,
)
from .api.protocols import BaseDialogManager, DialogManager, DialogProtocol
from .dialog import Dialog
from .manager.setup import setup_dialogs
from .manager.sub_manager import SubManager
from .window import Window

__all__ = [
    "DEFAULT_STACK_ID",
    "Dialog",
    "Data",
    "ChatEvent",
    "LaunchMode",
    "StartMode",
    "BaseDialogManager",
    "DialogManager",
    "DialogProtocol",
    "setup_dialogs",
    "ShowMode",
    "SubManager",
    "Window",
]



from typing import Dict

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from .base import Text


class Progress(Text):
    def __init__(
            self,
            field: str,
            width: int = 10,
            filled="🟥",
            empty="⬜",
            when: WhenCondition = None,
    ):
        super().__init__(when)
        self.field = field
        self.width = width
        self.filled = filled
        self.empty = empty

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        if manager.is_preview():
            percent = 15
        else:
            percent = data.get(self.field)
        done = round((self.width * percent) / 100)
        rest = self.width - done

        return self.filled * done + self.empty * rest + f" {percent: 2.0f}%"



from copy import deepcopy
from datetime import datetime
from typing import Optional, Set
from uuid import uuid4

from aiogram import Bot
from aiogram.types import (
    Audio, CallbackQuery, Document, Message, PhotoSize, Video,
)

from aiogram_dialog.api.entities import MediaAttachment, NewMessage
from aiogram_dialog.api.protocols import MessageManagerProtocol


def file_id(media: MediaAttachment) -> str:
    return media.file_id or str(uuid4())


MEDIA_CLASSES = {
    "audio": lambda x: Audio(
        file_id=file_id(x), file_unique_id=file_id(x),
        duration=1024,
    ),
    "document": lambda x: Document(
        file_id=file_id(x), file_unique_id=file_id(x),
    ),
    "photo": lambda x: [PhotoSize(
        file_id=file_id(x), file_unique_id=file_id(x),
        width=1024, height=1024,
    )],
    "video": lambda x: Video(
        file_id=file_id(x), file_unique_id=file_id(x),
        width=1024, height=1024, duration=1024,
    ),
}


class MockMessageManager(MessageManagerProtocol):
    def __init__(self):
        self.answered_callbacks: Set[str] = set()
        self.sent_messages = []
        self.last_message_id = 0

    def reset_history(self):
        self.sent_messages.clear()
        self.answered_callbacks.clear()

    def assert_one_message(self) -> None:
        assert len(self.sent_messages) == 1

    def last_message(self) -> Message:
        return self.sent_messages[-1]

    def first_message(self) -> Message:
        return self.sent_messages[0]

    def one_message(self) -> Message:
        self.assert_one_message()
        return self.first_message()

    async def remove_kbd(
            self, bot: Bot, old_message: Optional[Message],
    ) -> Optional[Message]:
        if not old_message:
            return
        data = old_message.model_dump()
        data["reply_markup"] = None
        message = Message(**data)
        self.sent_messages.append(message)
        return message

    async def answer_callback(
            self, bot: Bot, callback_query: CallbackQuery,
    ) -> None:
        self.answered_callbacks.add(callback_query.id)

    def assert_answered(self, callback_id: str) -> None:
        assert callback_id in self.answered_callbacks

    async def show_message(self, bot: Bot, new_message: NewMessage,
                           old_message: Optional[Message]) -> Message:
        message_id = self.last_message_id + 1
        self.last_message_id = message_id

        if new_message.media:
            contents = {
                "caption": new_message.text,
                new_message.media.type: MEDIA_CLASSES[new_message.media.type](
                    new_message.media,
                ),
            }
        else:
            contents = {
                "text": new_message.text,
            }

        message = Message(
            message_id=message_id,
            date=datetime.now(),
            chat=new_message.chat,
            reply_markup=deepcopy(new_message.reply_markup),
            **contents,
        )
        self.sent_messages.append(message)
        return message



from typing import Dict

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import (
    BaseScroll, OnPageChangedVariants, WhenCondition,
)
from .base import Text


class ScrollingText(Text, BaseScroll):
    def __init__(
            self,
            text: Text,
            id: str,
            page_size: int = 0,
            when: WhenCondition = None,
            on_page_changed: OnPageChangedVariants = None,
    ):
        Text.__init__(self, when=when)
        BaseScroll.__init__(self, id=id, on_page_changed=on_page_changed)
        self.text = text
        self.page_size = page_size

    def _get_page_count(
            self,
            text: str,
    ) -> int:
        return len(text) // self.page_size + bool(len(text) % self.page_size)

    async def _render_contents(
            self,
            data: Dict,
            manager: DialogManager,
    ) -> str:
        return await self.text.render_text(data, manager)

    async def _render_text(self, data, manager: DialogManager) -> str:
        text = await self._render_contents(data, manager)
        pages = self._get_page_count(text)
        page = await self.get_page(manager)
        last_page = pages - 1
        current_page = min(last_page, page)
        page_offset = current_page * self.page_size

        return text[page_offset: page_offset + self.page_size]

    async def get_page_count(self, data: Dict, manager: DialogManager) -> int:
        text = await self._render_contents(data, manager)
        return self._get_page_count(text)



from datetime import datetime
from logging import getLogger
from typing import Any, Dict, Optional

from aiogram import Router
from aiogram.fsm.state import State
from aiogram.types import CallbackQuery, Chat, Document, Message, User

from aiogram_dialog.api.entities import (
    ChatEvent, Context, Data, DEFAULT_STACK_ID, LaunchMode, NewMessage,
    ShowMode, Stack, StartMode,
)
from aiogram_dialog.api.exceptions import (
    IncorrectBackgroundError, NoContextError,
)
from aiogram_dialog.api.internal import (
    CONTEXT_KEY, STACK_KEY, STORAGE_KEY,
)
from aiogram_dialog.api.internal import (
    FakeChat, FakeUser,
)
from aiogram_dialog.api.protocols import (
    BaseDialogManager, DialogManager, DialogProtocol, DialogRegistryProtocol,
    MediaIdStorageProtocol, MessageManagerProtocol, MessageNotModified,
)
from aiogram_dialog.context.storage import StorageProxy
from aiogram_dialog.utils import get_media_id
from .bg_manager import BgManager

logger = getLogger(__name__)


class ManagerImpl(DialogManager):

    def __init__(
            self, event: ChatEvent,
            message_manager: MessageManagerProtocol,
            media_id_storage: MediaIdStorageProtocol,
            registry: DialogRegistryProtocol,
            router: Router,
            data: Dict,
    ):
        self.disabled = False
        self.message_manager = message_manager
        self.media_id_storage = media_id_storage
        self._event = event
        self._data = data
        self._show_mode: ShowMode = ShowMode.AUTO
        self._registry = registry
        self._router = router

    @property
    def show_mode(self) -> ShowMode:
        """Get current show mode, used for next show action."""
        return self._show_mode

    @show_mode.setter
    def show_mode(self, show_mode: ShowMode) -> None:
        """Set current show mode, used for next show action."""
        self._show_mode = show_mode

    @property
    def event(self) -> ChatEvent:
        return self._event

    @property
    def middleware_data(self) -> Dict:
        """Middleware data."""
        return self._data

    @property
    def dialog_data(self) -> Dict:
        """Dialog data for current context."""
        return self.current_context().dialog_data

    @property
    def start_data(self) -> Dict:
        """Start data for current context."""
        return self.current_context().start_data

    def check_disabled(self):
        if self.disabled:
            raise IncorrectBackgroundError(
                "Detected background access to dialog manager. "
                "Please use background manager available via `manager.bg()` "
                "method to access methods from background tasks",
            )

    async def load_data(self) -> Dict:
        context = self.current_context()
        return {
            "dialog_data": context.dialog_data,
            "start_data": context.start_data,
            "middleware_data": self._data,
            "event": self.event,
        }

    def is_preview(self) -> bool:
        return False

    def dialog(self) -> DialogProtocol:
        self.check_disabled()
        current = self.current_context()
        if not current:
            raise RuntimeError
        return self._registry.find_dialog(current.state)

    def current_context(self) -> Context:
        self.check_disabled()
        context = self._current_context_unsafe()
        if not context:
            logger.warning(
                "Trying to access current context, while no dialog is opened",
            )
            raise NoContextError
        return context

    def _current_context_unsafe(self) -> Optional[Context]:
        return self._data[CONTEXT_KEY]

    def has_context(self) -> bool:
        self.check_disabled()
        return bool(self._current_context_unsafe())

    def current_stack(self) -> Stack:
        self.check_disabled()
        return self._data[STACK_KEY]

    def storage(self) -> StorageProxy:
        return self._data[STORAGE_KEY]

    async def _remove_kbd(self) -> None:
        if self.current_stack().last_message_id is None:
            return
        chat = self._data["event_chat"]
        bot = self._data["bot"]
        message = Message(
            chat=chat,
            message_id=self.current_stack().last_message_id,
            date=datetime.now(),
        )
        await self.message_manager.remove_kbd(
            bot,
            message,
        )
        self.current_stack().last_message_id = None

    async def _process_last_dialog_result(
            self,
            start_data: Data,
            result: Any,
    ) -> None:
        """Process closing last dialog in stack."""
        await self._remove_kbd()

    async def done(self, result: Any = None) -> None:
        self.check_disabled()
        await self.dialog().process_close(result, self)
        old_context = self.current_context()
        await self.mark_closed()
        context = self._current_context_unsafe()
        if not context:
            await self._process_last_dialog_result(
                old_context.start_data,
                result,
            )
            return
        dialog = self.dialog()
        await dialog.process_result(old_context.start_data, result, self)
        new_context = self._current_context_unsafe()
        if new_context and context.id == new_context.id:
            await self.show()

    async def answer_callback(self) -> None:
        if not isinstance(self.event, CallbackQuery):
            return
        return await self.message_manager.answer_callback(
            bot=self._data["bot"],
            callback_query=self.event,
        )

    async def mark_closed(self) -> None:
        self.check_disabled()
        storage = self.storage()
        stack = self.current_stack()
        await storage.remove_context(stack.pop())
        if stack.empty():
            self._data[CONTEXT_KEY] = None
        else:
            intent_id = stack.last_intent_id()
            self._data[CONTEXT_KEY] = await storage.load_context(intent_id)
        await storage.save_stack(stack)

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
            show_mode: ShowMode = ShowMode.AUTO,
    ) -> None:
        self.check_disabled()
        self.show_mode = show_mode
        if mode is StartMode.NORMAL:
            await self._start_normal(state, data)
        elif mode is StartMode.RESET_STACK:
            await self.reset_stack(remove_keyboard=False)
            await self._start_normal(state, data)
        elif mode is StartMode.NEW_STACK:
            await self._start_new_stack(state, data)
        else:
            raise ValueError(f"Unknown start mode: {mode}")

    async def reset_stack(self, remove_keyboard: bool = True) -> None:
        self.check_disabled()
        storage = self.storage()
        stack = self.current_stack()
        while not stack.empty():
            await storage.remove_context(stack.pop())
        await storage.save_stack(stack)
        if remove_keyboard:
            await self._remove_kbd()
        self._data[CONTEXT_KEY] = None

    async def _start_new_stack(self, state: State, data: Data = None) -> None:
        stack = Stack()
        await self.bg(stack_id=stack.id).start(state, data, StartMode.NORMAL)

    async def _start_normal(self, state: State, data: Data = None) -> None:
        stack = self.current_stack()
        old_dialog: Optional[DialogProtocol] = None
        if not stack.empty():
            old_dialog = self.dialog()
            if old_dialog.launch_mode is LaunchMode.EXCLUSIVE:
                raise ValueError(
                    "Cannot start dialog on top "
                    "of one with launch_mode==SINGLE",
                )

        new_dialog = self._registry.find_dialog(state)
        launch_mode = new_dialog.launch_mode
        if launch_mode in (LaunchMode.EXCLUSIVE, LaunchMode.ROOT):
            await self.reset_stack(remove_keyboard=False)
        if launch_mode is LaunchMode.SINGLE_TOP:
            if new_dialog is old_dialog:
                await self.storage().remove_context(stack.pop())

        if self.has_context():
            await self.storage().save_context(self.current_context())
        context = stack.push(state, data)
        self._data[CONTEXT_KEY] = context
        await self.dialog().process_start(self, data, state)
        new_context = self._current_context_unsafe()
        if new_context and context.id == new_context.id:
            await self.show()

    async def next(self) -> None:
        context = self.current_context()
        states = self.dialog().states()
        current_index = states.index(context.state)
        new_state = states[current_index + 1]
        await self.switch_to(new_state)

    async def back(self) -> None:
        context = self.current_context()
        states = self.dialog().states()
        current_index = states.index(context.state)
        new_state = states[current_index - 1]
        await self.switch_to(new_state)

    async def switch_to(self, state: State) -> None:
        self.check_disabled()
        context = self.current_context()
        if context.state.group != state.group:
            raise ValueError(
                f"Cannot switch to another state group. "
                f"Current state: {context.state}, asked for {state}",
            )
        context.state = state

    async def show(self) -> None:
        stack = self.current_stack()
        bot = self._data["bot"]
        old_message = self._get_last_message()
        new_message = await self.dialog().render(self)
        if new_message.show_mode is ShowMode.AUTO:
            new_message.show_mode = self._calc_show_mode()
        await self._fix_cached_media_id(new_message)

        try:
            sent_message = await self.message_manager.show_message(
                bot, new_message, old_message,
            )
        except MessageNotModified:
            # nothing changed so nothing to save
            # we do not have the actual version of message
            logger.debug("MessageNotModified, not storing ids")
        else:
            self._save_last_message(sent_message)
            if new_message.media:
                await self.media_id_storage.save_media_id(
                    path=new_message.media.path,
                    url=new_message.media.url,
                    type=new_message.media.type,
                    media_id=get_media_id(sent_message),
                )
        self.show_mode = ShowMode.EDIT
        if isinstance(self.event, Message):
            stack.last_income_media_group_id = self.event.media_group_id

    async def _fix_cached_media_id(self, new_message: NewMessage):
        if not new_message.media or new_message.media.file_id:
            return
        new_message.media.file_id = await self.media_id_storage.get_media_id(
            path=new_message.media.path,
            url=new_message.media.url,
            type=new_message.media.type,
        )

    def _get_last_message(self) -> Optional[Message]:
        stack = self.current_stack()
        chat = self._data["event_chat"]
        if (
                isinstance(self.event, CallbackQuery) and
                self.event.message and
                stack.last_message_id == self.event.message.message_id
        ):
            return self.event.message
        if not stack or not stack.last_message_id:
            return None
        if stack.last_media_id:
            # we create document because
            # * there is no method to set content type explicitly
            # * we don't really care fo exact content type
            document = Document(
                file_id=stack.last_media_id,
                file_unique_id=stack.last_media_unique_id,
            )
            text = None
        else:
            document = None
            # we set some non empty-text which is not equal to anything
            text = "𝔞𝔦𝔬𝔤𝔯𝔞𝔪 𝔡𝔦𝔞𝔩𝔬𝔤 𝔲𝔫𝔦𝔮𝔲𝔢 𝔱𝔢𝔵𝔱"
        return Message(
            message_id=stack.last_message_id,
            document=document,
            text=text,
            chat=chat,
            date=datetime.now(),
        )

    def _save_last_message(self, message: Message):
        stack = self.current_stack()
        stack.last_message_id = message.message_id
        media_id = get_media_id(message)
        if media_id:
            stack.last_media_id = media_id.file_id
            stack.last_media_unique_id = media_id.file_unique_id
        else:
            stack.last_media_id = None
            stack.last_media_unique_id = None

    def _calc_show_mode(self) -> ShowMode:
        if self.show_mode is not ShowMode.AUTO:
            return self.show_mode
        if self.current_stack().id != DEFAULT_STACK_ID:
            return ShowMode.EDIT
        if isinstance(self.event, Message):
            if self.event.media_group_id is None:
                return ShowMode.SEND
            elif self.event.media_group_id == \
                    self.current_stack().last_income_media_group_id:
                return ShowMode.EDIT
            else:
                return ShowMode.SEND
        return ShowMode.EDIT

    async def update(self, data: Dict) -> None:
        self.current_context().dialog_data.update(data)
        await self.show()

    def find(self, widget_id) -> Optional[Any]:
        widget = self.dialog().find(widget_id)
        if not widget:
            return None
        return widget.managed(self)

    def is_same_chat(self, user: User, chat: Chat) -> bool:
        if "event_chat" not in self._data:
            return False

        current_chat = self._data["event_chat"]
        current_user = self.event.from_user
        return user.id == current_user.id and chat.id == current_chat.id

    def _get_fake_user(self, user_id: Optional[int] = None) -> User:
        """Get User if we have info about him or FakeUser instead."""
        current_user = self.event.from_user
        if user_id in (None, current_user.id):
            return current_user
        return FakeUser(id=user_id, is_bot=False, first_name="")

    def _get_fake_chat(self, chat_id: Optional[int] = None) -> Chat:
        """Get Chat if we have info about him or FakeChat instead."""
        if "event_chat" in self._data:
            current_chat = self._data["event_chat"]
            if chat_id in (None, current_chat.id):
                return current_chat
        elif chat_id is None:
            raise ValueError(
                "Explicit `chat_id` is required "
                "for events without current chat",
            )
        return FakeChat(id=chat_id, type="")

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> BaseDialogManager:
        user = self._get_fake_user(user_id)
        chat = self._get_fake_chat(chat_id)
        intent_id = None
        if stack_id is None:
            if self.is_same_chat(user, chat):
                stack_id = self.current_stack().id
                if self.has_context():
                    intent_id = self.current_context().id
            else:
                stack_id = DEFAULT_STACK_ID

        return BgManager(
            user=user,
            chat=chat,
            bot=self._data["bot"],
            router=self._router,
            intent_id=intent_id,
            stack_id=stack_id,
            load=load,
        )

    async def close_manager(self) -> None:
        self.check_disabled()
        self.disabled = True
        del self.media_id_storage
        del self.message_manager
        del self._event
        del self._data



from logging import getLogger
from typing import Any, Dict, Optional

from aiogram import Bot, Router
from aiogram.fsm.state import State
from aiogram.types import Chat, User

from aiogram_dialog.api.entities import (
    Data,
    DEFAULT_STACK_ID,
    DialogAction,
    DialogStartEvent,
    DialogSwitchEvent,
    DialogUpdate,
    DialogUpdateEvent,
    ShowMode,
    StartMode,
)
from aiogram_dialog.api.internal import (
    FakeChat, FakeUser,
)
from aiogram_dialog.api.protocols import BaseDialogManager, BgManagerFactory
from aiogram_dialog.manager.updater import Updater
from aiogram_dialog.utils import is_chat_loaded, is_user_loaded

logger = getLogger(__name__)


class BgManager(BaseDialogManager):
    def __init__(
            self,
            user: User,
            chat: Chat,
            bot: Bot,
            router: Router,
            intent_id: Optional[str],
            stack_id: Optional[str],
            load: bool = False,
    ):
        self.user = user
        self.chat = chat
        self.bot = bot
        self._router = router
        self._updater = Updater(router)
        self.intent_id = intent_id
        self.stack_id = stack_id
        self.load = load

    def bg(
            self,
            user_id: Optional[int] = None,
            chat_id: Optional[int] = None,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> "BaseDialogManager":
        if chat_id in (None, self.chat.id):
            chat = self.chat
        else:
            chat = FakeChat(id=chat_id, type="")

        if user_id in (None, self.user.id):
            user = self.user
        else:
            user = FakeUser(id=user_id, is_bot=False, first_name="")

        same_chat = user.id == self.user.id and chat.id == self.chat.id
        if stack_id is None:
            if same_chat:
                stack_id = self.stack_id
                intent_id = self.intent_id
            else:
                stack_id = DEFAULT_STACK_ID
                intent_id = None
        else:
            intent_id = None

        return BgManager(
            user=user,
            chat=chat,
            bot=self.bot,
            router=self._router,
            intent_id=intent_id,
            stack_id=stack_id,
            load=load,
        )

    def _base_event_params(self):
        return {
            "from_user": self.user,
            "chat": self.chat,
            "intent_id": self.intent_id,
            "stack_id": self.stack_id,
        }

    async def _notify(self, event: DialogUpdateEvent):
        await self._updater.notify(
            bot=self.bot, update=DialogUpdate(aiogd_update=event),
        )

    async def _load(self):
        if self.load:
            if not is_chat_loaded(self.chat):
                logger.debug("load chat: %s", self.chat.id)
                self.chat = await self.bot.get_chat(self.chat.id)
            if not is_user_loaded(self.user):
                logger.debug(
                    "load user %s from chat %s", self.chat.id, self.user.id,
                )
                chat_member = await self.bot.get_chat_member(
                    self.chat.id, self.user.id,
                )
                self.user = chat_member.user

    async def done(self, result: Any = None) -> None:
        await self._load()
        await self._notify(
            DialogUpdateEvent(
                action=DialogAction.DONE, data=result,
                **self._base_event_params(),
            ),
        )

    async def start(
            self,
            state: State,
            data: Data = None,
            mode: StartMode = StartMode.NORMAL,
            show_mode: ShowMode = ShowMode.AUTO,
    ) -> None:
        await self._load()
        await self._notify(
            DialogStartEvent(
                action=DialogAction.START,
                data=data,
                new_state=state,
                mode=mode,
                show_mode=show_mode,
                **self._base_event_params(),
            ),
        )

    async def switch_to(self, state: State) -> None:
        await self._load()
        await self._notify(
            DialogSwitchEvent(
                action=DialogAction.SWITCH,
                data={},
                new_state=state,
                **self._base_event_params(),
            ),
        )

    async def update(self, data: Dict) -> None:
        await self._load()
        await self._notify(
            DialogUpdateEvent(
                action=DialogAction.UPDATE, data=data,
                **self._base_event_params(),
            ),
        )


class BgManagerFactoryImpl(BgManagerFactory):
    def __init__(self, router: Router):
        self._router = router

    def bg(
            self,
            bot: Bot,
            user_id: int,
            chat_id: int,
            stack_id: Optional[str] = None,
            load: bool = False,
    ) -> "BaseDialogManager":
        chat = FakeChat(id=chat_id, type="")
        user = FakeUser(id=user_id, is_bot=False, first_name="")
        if stack_id is None:
            stack_id = DEFAULT_STACK_ID

        return BgManager(
            user=user,
            chat=chat,
            bot=bot,
            router=self._router,
            intent_id=None,
            stack_id=stack_id,
            load=load,
        )



from abc import abstractmethod
from typing import Any, Callable, Union

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.protocols import DialogManager


class WidgetEventProcessor:
    @abstractmethod
    async def process_event(
            self,
            event: ChatEvent,
            source: Any,
            manager: DialogManager,
            *args,
            **kwargs,
    ):
        raise NotImplementedError


class SimpleEventProcessor(WidgetEventProcessor):
    def __init__(self, callback: Callable):
        self.callback = callback

    async def process_event(
            self,
            event: ChatEvent,
            source: Any,
            manager: DialogManager,
            *args,
            **kwargs,
    ):
        if self.callback:
            await self.callback(event, source, manager, *args, **kwargs)


def ensure_event_processor(
        processor: Union[Callable, WidgetEventProcessor, None],
) -> WidgetEventProcessor:
    if isinstance(processor, WidgetEventProcessor):
        return processor
    else:
        return SimpleEventProcessor(processor)



import pytest
from aiogram import Dispatcher
from aiogram.filters import CommandStart
from aiogram.fsm.state import State, StatesGroup
from aiogram.types import Message

from aiogram_dialog import (
    Dialog, DialogManager, setup_dialogs, StartMode, Window,
)
from aiogram_dialog.test_tools import BotClient, MockMessageManager
from aiogram_dialog.test_tools.keyboard import InlineButtonTextLocator
from aiogram_dialog.widgets.kbd import Back, Cancel, Next, Start
from aiogram_dialog.widgets.text import Const, Format


class MainSG(StatesGroup):
    start = State()
    next = State()


class SecondarySG(StatesGroup):
    start = State()


async def start(message: Message, dialog_manager: DialogManager):
    await dialog_manager.start(MainSG.start, mode=StartMode.RESET_STACK)


@pytest.fixture()
def message_manager() -> MockMessageManager:
    return MockMessageManager()


@pytest.fixture()
def client(dp) -> BotClient:
    return BotClient(dp)


@pytest.fixture()
def dp(message_manager: MockMessageManager):
    dp = Dispatcher()
    dp.message.register(start, CommandStart())

    dp.include_router(Dialog(
        Window(
            Const("First"),
            Next(),
            Start(Const("Start"), state=SecondarySG.start, id="start"),
            Cancel(),
            state=MainSG.start,
        ),
        Window(
            Const("Second"),
            Back(),
            state=MainSG.next,
        ),
    ))
    dp.include_router(Dialog(
        Window(
            Format("Subdialog"),
            Cancel(),
            state=SecondarySG.start,
        ),
    ))
    setup_dialogs(dp, message_manager=message_manager)
    return dp


@pytest.mark.asyncio
async def test_start(dp, message_manager, client):
    # start
    await client.send("/start")
    first_message = message_manager.one_message()
    assert first_message.text == "First"
    assert first_message.reply_markup


@pytest.mark.asyncio
async def test_next_back(dp, message_manager, client):
    await client.send("/start")
    first_message = message_manager.one_message()

    # click next
    message_manager.reset_history()
    callback_id = await client.click(
        first_message, InlineButtonTextLocator("Next"),
    )
    message_manager.assert_answered(callback_id)
    second_message = message_manager.one_message()
    assert second_message.text == "Second"

    # click back
    message_manager.reset_history()
    callback_id = await client.click(
        second_message, InlineButtonTextLocator("Back"),
    )
    message_manager.assert_answered(callback_id)
    last_message = message_manager.one_message()
    assert last_message.text == "First"
    assert last_message.reply_markup


@pytest.mark.asyncio
async def test_finish_last(dp, message_manager, client):
    await client.send("/start")
    first_message = message_manager.one_message()

    # click back
    message_manager.reset_history()
    callback_id = await client.click(
        first_message, InlineButtonTextLocator("Cancel"),
    )
    message_manager.assert_answered(callback_id)
    last_message = message_manager.one_message()
    assert not last_message.reply_markup, "Keyboard closed"


@pytest.mark.asyncio
async def test_reset_stack(dp, message_manager, client):
    for _ in range(200):
        message_manager.reset_history()
        await client.send("/start")
        first_message = message_manager.one_message()
        assert first_message.text == "First"

    message_manager.reset_history()
    callback_id = await client.click(
        first_message, InlineButtonTextLocator("Cancel"),
    )
    message_manager.assert_answered(callback_id)
    last_message = message_manager.one_message()
    assert not last_message.reply_markup, "Keyboard closed"


@pytest.mark.asyncio
async def test_subdialog(dp, message_manager, client):
    await client.send("/start")
    first_message = message_manager.one_message()

    # start subdialog
    message_manager.reset_history()
    callback_id = await client.click(
        first_message, InlineButtonTextLocator("Start"),
    )
    message_manager.assert_answered(callback_id)
    second_message = message_manager.one_message()
    assert second_message.text == "Subdialog"

    # close subdialog
    message_manager.reset_history()
    callback_id = await client.click(
        second_message, InlineButtonTextLocator("Cancel"),
    )
    message_manager.assert_answered(callback_id)
    last_message = message_manager.one_message()
    assert last_message.text == "First"
    assert last_message.reply_markup



from dataclasses import dataclass
from typing import Optional

from aiogram.types import Chat, InlineKeyboardMarkup

from aiogram_dialog.api.entities import MediaAttachment, ShowMode


@dataclass
class NewMessage:
    chat: Chat
    text: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    parse_mode: Optional[str] = None
    show_mode: ShowMode = ShowMode.AUTO
    disable_web_page_preview: Optional[bool] = None
    media: Optional[MediaAttachment] = None



from logging import getLogger
from typing import Optional, Tuple

from aiogram.types import (
    CallbackQuery, Chat, ChatMemberUpdated, Message, User,
)

from aiogram_dialog.api.entities import (
    ChatEvent, DialogUpdateEvent, MediaId, NewMessage,
)

logger = getLogger(__name__)

CB_SEP = "\x1D"


def get_chat(event: ChatEvent) -> Chat:
    if isinstance(event, (Message, DialogUpdateEvent, ChatMemberUpdated)):
        return event.chat
    elif isinstance(event, CallbackQuery):
        if not event.message:
            return Chat(id=event.from_user.id, type="")
        return event.message.chat


def is_chat_loaded(chat: Chat) -> bool:
    """
    Check if chat is correctly loaded from telegram.

    For internal events it can be created with no data inside as a FakeChat
    """
    return getattr(chat, "fake", False)


def is_user_loaded(user: User) -> bool:
    """
    Check if user is correctly loaded from telegram.

    For internal events it can be created with no data inside as a FakeUser
    """
    return getattr(user, "fake", False)


def get_media_id(message: Message) -> Optional[MediaId]:
    media = (
        message.audio or
        message.animation or
        message.document or
        (message.photo[-1] if message.photo else None) or
        message.video
    )
    if not media:
        return None
    return MediaId(
        file_id=media.file_id,
        file_unique_id=media.file_unique_id,
    )


def intent_callback_data(
        intent_id: str, callback_data: Optional[str],
) -> Optional[str]:
    if callback_data is None:
        return None
    prefix = intent_id + CB_SEP
    if callback_data.startswith(prefix):
        return callback_data
    return prefix + callback_data


def add_indent_id(message: NewMessage, intent_id: str):
    if not message.reply_markup:
        return
    for row in message.reply_markup.inline_keyboard:
        for button in row:
            button.callback_data = intent_callback_data(
                intent_id, button.callback_data,
            )


def remove_indent_id(callback_data: str) -> Tuple[Optional[str], str]:
    if CB_SEP in callback_data:
        intent_id, new_data = callback_data.split(CB_SEP, maxsplit=1)
        return intent_id, new_data
    return None, callback_data



from operator import itemgetter
from typing import Callable, Dict, Sequence, Union

from magic_filter import MagicFilter

ItemsGetter = Callable[[Dict], Sequence]
ItemsGetterVariant = Union[str, ItemsGetter, MagicFilter, Sequence]


def _get_identity(items: Sequence) -> ItemsGetter:
    def identity(data) -> Sequence:
        return items

    return identity


def _get_magic_getter(f: MagicFilter) -> ItemsGetter:
    def items_magic(data: Dict) -> Sequence:
        items = f.resolve(data)
        if isinstance(items, Sequence):
            return items
        else:
            return []

    return items_magic


def get_items_getter(attr_val: ItemsGetterVariant) -> ItemsGetter:
    if isinstance(attr_val, str):
        return itemgetter(attr_val)
    elif isinstance(attr_val, MagicFilter):
        return _get_magic_getter(attr_val)
    elif isinstance(attr_val, Callable):
        return attr_val
    else:
        return _get_identity(attr_val)



from enum import Enum


class LaunchMode(Enum):
    """
    Modes of launching new dialog.

    `ROOT` dialogs will be always a root dialog in stack.
        Starting such dialogs will automatically reset stack.
        Example: main menu

    `EXCLUSIVE` dialogs can be only a single dialog in stack.
        Starting such dialogs will automatically reset stack.
        Starting other dialogs on top of them is forbidden
        Example: Banners

    `SINGLE_TOP` dialogs will not be repeated on top of stack.
        Starting the same dialog right on top of it will just replace it.
        Example: product page

    `STANDARD` dialogs have no limitations themselves
    """

    STANDARD = "standard"
    ROOT = "root"
    EXCLUSIVE = "exclusive"
    SINGLE_TOP = "single_top"



from abc import abstractmethod
from typing import Dict, Protocol

from aiogram import Router

from aiogram_dialog.api.entities import ChatEvent
from aiogram_dialog.api.protocols import (
    DialogManager, DialogRegistryProtocol,
)


class DialogManagerFactory(Protocol):
    @abstractmethod
    def __call__(
            self, event: ChatEvent, data: Dict,
            registry: DialogRegistryProtocol,
            router: Router,
    ) -> DialogManager:
        raise NotImplementedError



from abc import abstractmethod
from typing import Optional, Protocol

from aiogram.types import ContentType

from aiogram_dialog.api.entities import MediaId


class MediaIdStorageProtocol(Protocol):
    @abstractmethod
    async def get_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
    ) -> Optional[MediaId]:
        raise NotImplementedError

    @abstractmethod
    async def save_media_id(
            self,
            path: Optional[str],
            url: Optional[str],
            type: ContentType,
            media_id: MediaId,
    ) -> None:
        raise NotImplementedError



import re
from typing import Optional, Protocol

from aiogram.types import InlineKeyboardButton, Message


class InlineButtonLocator(Protocol):
    def find_button(
            self, message: Message,
    ) -> Optional[InlineKeyboardButton]:
        raise NotImplementedError


class InlineButtonTextLocator:
    def __init__(self, regex: str):
        self.regex = re.compile(regex)

    def find_button(
            self, message: Message,
    ) -> Optional[InlineKeyboardButton]:
        if not message.reply_markup:
            return None
        for row in message.reply_markup.inline_keyboard:
            for button in row:
                if self.regex.fullmatch(button.text):
                    return button
        return None

    def __repr__(self):
        return f"InlineButtonTextLocator({self.regex.pattern!r})"


class InlineButtonPositionLocator:
    def __init__(self, row: int, column: int):
        self.row = row
        self.column = column

    def find_button(
            self, message: Message,
    ) -> Optional[InlineKeyboardButton]:
        if not message.reply_markup:
            return None
        try:
            return message.reply_markup.inline_keyboard[self.row][self.column]
        except IndexError:
            return None

    def __repr__(self):
        return f"InlineButtonPositionLocator" \
               f"(row={self.row}, column={self.column})"


class InlineButtonDataLocator:
    def __init__(self, regex: str):
        self.regex = re.compile(regex)

    def find_button(
            self, message: Message,
    ) -> Optional[InlineKeyboardButton]:
        if not message.reply_markup:
            return None
        for row in message.reply_markup.inline_keyboard:
            for button in row:
                if not button.callback_data:
                    continue
                if self.regex.fullmatch(button.callback_data):
                    return button
        return None

    def __repr__(self):
        return f"InlineButtonDataLocator({self.regex.pattern!r})"



"""
Dynamic media.

Originally developed in aiogram-dialog-extras project
https://github.com/SamWarden/aiogram_dialog_extras
"""

from operator import itemgetter
from typing import Callable, Optional, Union

from aiogram_dialog import DialogManager
from aiogram_dialog.api.entities import MediaAttachment
from aiogram_dialog.widgets.common import WhenCondition
from aiogram_dialog.widgets.media import Media

MediaSelector = Callable[[dict], MediaAttachment]


class DynamicMedia(Media):
    def __init__(
            self,
            selector: Union[str, MediaSelector],
            when: WhenCondition = None,
    ):
        super().__init__(when)
        if isinstance(selector, str):
            self.selector: MediaSelector = itemgetter(selector)
        else:
            self.selector = selector

    async def _render_media(
            self, data: dict, manager: DialogManager,
    ) -> Optional[MediaAttachment]:
        media: Optional[MediaAttachment] = self.selector(data)
        return media



from __future__ import annotations

from abc import abstractmethod
from typing import Dict, Protocol, Union

from magic_filter import MagicFilter

from aiogram_dialog.api.protocols import DialogManager


class Predicate(Protocol):
    @abstractmethod
    def __call__(
            self,
            data: Dict,
            widget: Whenable,
            dialog_manager: DialogManager,
    ) -> bool:
        """
        Check if widget should be shown.

        :param data: Data received from getter
        :param widget: Widget we are working with
        :param dialog_manager: Dialog manager to access current context
        :return: ``True`` if widget has to be shown, ``False`` otherwise
        """
        raise NotImplementedError


WhenCondition = Union[str, MagicFilter, Predicate, None]


def new_when_field(fieldname: str) -> Predicate:
    def when_field(
            data: Dict, widget: "Whenable", manager: DialogManager,
    ) -> bool:
        return bool(data.get(fieldname))

    return when_field


def new_when_magic(f: MagicFilter) -> Predicate:
    def when_magic(
            data: Dict, widget: "Whenable", manager: DialogManager,
    ) -> bool:
        return f.resolve(data)

    return when_magic


def true_condition(data: Dict, widget: "Whenable", manager: DialogManager):
    return True


class Whenable:
    def __init__(self, when: WhenCondition = None):
        self.condition: Predicate
        if when is None:
            self.condition = true_condition
        elif isinstance(when, str):
            self.condition = new_when_field(when)
        elif isinstance(when, MagicFilter):
            self.condition = new_when_magic(when)
        else:
            self.condition = when

    def is_(self, data: Dict, manager: DialogManager):
        return self.condition(data, self, manager)



from typing import Dict

from aiogram_dialog.api.protocols import DialogManager
from aiogram_dialog.widgets.common import WhenCondition
from .base import Text


class _FormatDataStub:
    def __init__(self, name="", data=None):
        self.name = name
        self.data = data or {}

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        if not self.name:
            return _FormatDataStub(item)
        return _FormatDataStub(f"{self.name}[{item}]")

    def __getattr__(self, item):
        return _FormatDataStub(f"{self.name}.{item}")

    def __format__(self, format_spec):
        if format_spec:
            res = f"{self.name}:{format_spec}"
        else:
            res = self.name
        return f"{{{res}}}"


class Format(Text):
    def __init__(self, text: str, when: WhenCondition = None):
        super().__init__(when=when)
        self.text = text

    async def _render_text(
            self, data: Dict, manager: DialogManager,
    ) -> str:
        if manager.is_preview():
            return self.text.format_map(_FormatDataStub(data=data))
        return self.text.format_map(data)



from datetime import datetime

from aiogram import Bot
from aiogram.types import (
    Chat, Message, Update, User,
)
from aiogram_dialog.api.entities.update_event import DialogUpdate


class FakeBot(Bot):
    def __init__(self):
        pass  # do not call super, so it is invalid bot, used only as a stub

    @property
    def id(self):
        return 1

    def __call__(self, *args, **kwargs) -> None:
        raise RuntimeError("Fake bot should not be used to call telegram")
    def __hash__(self) -> int:
        return 1

    def __eq__(self, other) -> bool:
        return self is other


def new_message(
        text: str,
):
    chat = Chat(id=1, type="private")
    user = User(
        id=1, is_bot=False,
        first_name=f"User_1",
    )
    return Message(
        message_id=1,
        date=datetime.fromtimestamp(1234567890),
        chat=chat,
        from_user=user,
        text=text,
    )


def test_upd():
    update = Update(
        update_id=1,
        message=new_message("a"),
        bot=FakeBot(),
    )
    print(update.event_type)

    update = Update(
        update_id=1,
        message=new_message("a"),
        bot=FakeBot(),
    )
    print(update.event_type)



from dataclasses import dataclass
from typing import Optional

from aiogram.types import ContentType


@dataclass
class MediaId:
    file_id: str
    file_unique_id: Optional[str] = None

    def __eq__(self, other):
        if type(other) is not MediaId:
            return False
        if self.file_unique_id is None or other.file_unique_id is None:
            return self.file_id == other.file_id
        return self.file_unique_id == other.file_unique_id


class MediaAttachment:
    def __init__(
            self,
            type: ContentType,
            url: Optional[str] = None,
            path: Optional[str] = None,
            file_id: Optional[MediaId] = None,
            use_pipe: bool = False,
            **kwargs,
    ):
        if not (url or path or file_id):
            raise ValueError("Neither url nor path not file_id are provided")
        self.type = type
        self.url = url
        self.path = path
        self.file_id = file_id
        self.use_pipe = use_pipe
        self.kwargs = kwargs
